import React from "react";

type StateArgs = {done?:boolean, open?:boolean};
type StateObj = {done:boolean, open:boolean};
type StateBinding = [state:StateObj, update:(args:StateArgs)=>void];

const CTX = React.createContext([{done:true, open:false}, (args)=>{}] as StateBinding);

export const Group =(props:{children:React.JSX.Element|React.JSX.Element[]})=>
{
    const [stateGet, stateSet] = React.useState({done:true, open:false} as StateObj);
    return <CTX.Provider value={[stateGet, (args:StateArgs)=> stateSet({...stateGet, ...args})]}>{props.children}</CTX.Provider>;
};

export const Menu =(props:{children:React.JSX.Element|React.JSX.Element[]})=>
{
    const [stateGet, stateSet] = React.useContext(CTX);
    const refElement:React.MutableRefObject<HTMLElement|null> = React.useRef( null );
    const refControl:React.MutableRefObject<CollapseControls|null> = React.useRef( null );

    type MenuClassStates = {Keep:string, Open:string, Shut:string, Move:string, Exit:string};
    const Classes:MenuClassStates =
    {
      Keep: "relative transition-all border(8 black) top-0",
      Open: "h-auto w-full    top-0 duration-700",
      Shut: "h-0    w-[300px] top-0  duration-300",
      Move: "",
      Exit: "h-auto w-auto top-36 px-4 opacity-0"
    };
    const Window = window as {TwindInst?:(c:string)=>string};
    if(Window.TwindInst)
    {
      for(let stateName in Classes)
      {
        Classes[stateName as keyof MenuClassStates] = Window.TwindInst(Classes[stateName as keyof MenuClassStates]);
      }
    }

    React.useEffect(()=>refControl.current = Collapser(refElement.current as HTMLElement, true), []);
    React.useEffect(()=> {refControl.current && refControl.current(stateGet.open, 1000, ()=>stateSet({done:true}))}, [stateGet.open])

    return <div class={Classes.Keep} data-class-keep={Classes.Keep} data-class-open={Classes.Open} data-class-shut={Classes.Shut} ref={refElement as React.Ref<HTMLDivElement>}>
      { (!stateGet.open && stateGet.done) ? null : props.children}
    </div>;
};

export const Button =()=>
{
    const [stateGet, stateSet] = React.useContext(CTX);
    return <>
        <p>{JSON.stringify(stateGet)}</p>
        <button class="px-10 py-2 bg-red-500 text-white" onClick={e=>stateSet({open:true,  done:false})}>Open</button>
        <button class="px-10 py-2 bg-red-500 text-white" onClick={e=>stateSet({open:false, done:false})}>Close</button>
    </>;
};

type DoneCallback =(openState:boolean)=>void;
export type CollapseControls =(inOpen?:boolean, inMs?:number, inDone?:DoneCallback)=>void;
export function Collapser(inElement:HTMLElement, initialState = false)
{
    let userDone:DoneCallback = (openState) => {};
    let userMode = initialState;
    let frameRequest = 0;
    let inTransition = false;

    let transitions = new Set();
    let measurements = false;

    const run = (inEvent:TransitionEvent)=>
    {
      if (inEvent.target == inElement)
      {
        transitions.add(inEvent.propertyName);
      }
    };

    const end = (inEvent:TransitionEvent)=>
    {
      if (inEvent.target == inElement)
      {
        transitions.delete(inEvent.propertyName);
        if(transitions.size === 0)
        {
          if (userMode)
          {
            //inElement.setAttribute("style", "");
          }
          inTransition = false;
          userDone(userMode);          
        }

      }
    };

    const getMeasurements =(inClasses:Record<string, string>)=>
    {
      inElement.setAttribute("style", `transition: none;`);
      const initialClasses = inElement.getAttribute("class")||"";
      const output = {} as Record<string, {width:number, height:number}>;
      Object.entries(inClasses).forEach(([key, value])=>
      {
        inElement.setAttribute("class", value);
        console.log(inElement.offsetHeight, inElement.clientHeight);
        output[key] = { width: inElement.offsetWidth, height: inElement.offsetHeight };
      });
      inElement.removeAttribute("style");
      inElement.setAttribute("class", initialClasses);
      return output;
    };
  
    inElement.addEventListener("transitionend", end);
    inElement.addEventListener("transitionrun", run);

    return function(inOpen, inMs, inDone)
    {
      cancelAnimationFrame(frameRequest);

      if(arguments.length)
      {
        userDone = inDone|| ((m)=>{}) as DoneCallback;
        userMode = inOpen === true;

        if((inOpen || !measurements) && !inTransition)
        {
          measurements = getMeasurements({
            Open: inElement.getAttribute("data-class-open")||"",
            Shut: inElement.getAttribute("data-class-shut")||"",
          });
          console.log(measurements);
        }

        inElement.style.width = inElement.offsetWidth + "px";
        inElement.style.height = inElement.offsetHeight + "px";
        inElement.style.overflow = "hidden";

        inTransition = true;

        frameRequest = requestAnimationFrame(()=>
        {
          const measure = inOpen ? measurements.Open : measurements.Shut;
          inElement.style.width = measure.width + "px";
          inElement.style.height = measure.height + "px";
          inElement.className = inElement.getAttribute("data-class-keep") + " " + inElement.getAttribute(`data-class-${inOpen?"open":"shut"}`);
          
        });
      }
      else
      {
        inElement.removeEventListener("transitionrun", run);
        inElement.removeEventListener("transitionend", end);
      }
    } as CollapseControls;
}